#include <curses.h> 
#include <sys/types.h> 
#include <sys/time.h>
#include <errno.h>
#include <math.h>
#include "score.h"
#include "fishmonger.h"
#ifdef SYSV
#include <stropts.h>
#include <poll.h>
#define bzero(s,l)	memset((s),0,(l))
#define random rand
#define srandom srand
#endif
#include <sys/ioctl.h>

#define BEEP_CHARS           ""
#define TIMEOUT              5

#define DEFSCORES            18
#define SCOREFILE            ".fmscores"

#define ELSEWHERE            30000

#define ON                   1
#define OFF                  0

#define LEFT                 1
#define RIGHT                0

#define SHORT_TERM           1
#define LONG_TERM            0

#define LASER_BEAM_CHAR      '~'
#define NUM_LIVES            3
#define NUM_SMART_BOMBS      2
#define NEW_SMART_BOMB       5000

#define NUM_HUMANOIDS        10
#define NUM_LANDERS          10
#define NUM_PODS             2
#define NUM_SWARMERS         6
#define NUM_BOMBERS          2

#define HUMANOID_FREE        -1

#define BOMBER_DELAY         5
#define BOMBER_FOLLOW_DELAY  2
#define BOMB_FREQUENCY       15

#define LANDER_IDLE          -1
#define LANDER_STEALING      -2
#define LANDER_MATCH_HEIGHT  40
#define LANDER_DECIDE_STEAL  2000
#define NEW_WAVE_STEAL_RATE_CHANGE    300
#define LANDER_VERTICAL_RATE 15
#define LANDER_FIRE_RATE     20

#define BAITER_TURNAROUND    35

#define NUM_COLUMNS          80
#define BOTTOM               22
#define TOP                  6
#define STARTING_X           15
#define STARTING_Y           15
#define LEFT_X_POSITION      STARTING_X
#define RIGHT_X_POSITION     NUM_COLUMNS-STARTING_X-4
#define WINDOW_SLIDE_FACTOR  20
#define MAX_THRUST           4
#define SPACE_RANGE          800

#define NEW_WAVE             15
#define NO_MORE_ACTIVITY     30
#define NEW_ALIEN_RATE       400
#define NEW_WAVE_ALIEN_RATE_CHANGE    80
#define POD_CHANCE           2
#define BOMBER_CHANCE        5
#define BAITER_CHANCE        25


#define EXPLOSION_DELAY      90000
#define SMALL_DELAY          10000
#define MATERIALIZE_DELAY    6000

#define NUM_ZOOMS            30

#define ACTIVE_ARRAY_SIZE    400

#define UNUSED               -1
#define LANDER               0
#define MUTANT               1
#define BAITER               2
#define POD                  3
#define BOMBER               4
#define SWARMER              5
#define HUMANOID             6
#define BOMB                 7
#define MISSILE              8
#define DEBRIS1              9
#define DEBRIS2              10

#define DEBRIS_BITS          4
#define SMALL_EXPLOSION_SIZE 4

#define THRUST_DECAY	     6

#ifdef DO_GROUND
#	define ISBACK(ch)  ((ch)==' ' || (ch)=='\\' || (ch)=='/' || (ch)=='_')
#	define NUM_GROUND  100
#	define MAX_GROUND  5
#	define GROUND_DIV  4

int	ground_height[NUM_GROUND];
int	ground_slope[NUM_GROUND];
int	ground = 1;			/* zero if no ground please */

#else /* DO_GROUND */
#	define ISBACK(ch)	((ch)==' ')
#endif /* DO_GROUND */

/***************************************************************************\
|*									   *|
|*  I'm gonna define the standard movement keys here at the top of the	   *|
|*  file so that you can see exactly what I've done.  Other than what	   *|
|*  follows here, and the getenv() bit in main(), the only other change	   *|
|*  is to the key-recognition near the bottom of main() - Mirk		   *|
|*									   *|
\***************************************************************************/

char _up_key = 'q';
char _down_key = 'a';
char _right_key = '.';
char _left_key = ',';
char _fire_key = ' ';
char _smart_bomb_key = 's';
char _hyperspace_key = 'w';
char _suicide_key = '*';

/****************************************************************************\
  GLOBAL OSTEOPATHS, (whoops, I mean variables)

\****************************************************************************/

extern time_t timeofday;

int score, lives, smart_bombs, ship_x, ship_y, facing, hit, home_x, thrust;

int lander_decide_steal, num_humanoids, num_landers, new_alien_rate;

int next_smart_bomb;
int thrust_timer;

time_t *big;

char ship[2][5] = { ">8=-", "-=8<" };

char pieces[2][9] = { ">>88==--", "--==88<<" };

char things[11][4] = { ">o<", "}X{", "-=-", "()", "%", "*", "P", "x", ".", "-", "`"  };
    char blank_things[11][4] = { "   ", "   ", "   ", "  ", " ", " ", " ", " ", " ", " ", " " };
    
    ALIEN active[ACTIVE_ARRAY_SIZE];

    int array_top;
    
#ifdef SYSV
#if 0
void usleep (us)
unsigned us;
{
	(void) nap (us / 1000);
}
#endif
/*
 * Assumptions:  Only ever select with nfds == 0, or rfd == 0, or
 *		 rfd on file descriptor 0 (stdin).
 *		 Timeout NULL implies blocking select.
 */
int sysv_select (nfds, rfd, wfd, xfd, timeout)
int nfds;
fd_set *rfd, *wfd, *xfd;
struct timeval *timeout;
{
	struct pollfd pollrfd;
	int ret, time;
	unsigned long pollnfds = ((nfds == 0 || rfd == 0) ? 0L : 1L);
	
	if ( timeout )
		time = (timeout->tv_sec * 1000) + (timeout->tv_usec / 1000);
	else
		time = -1;
		
	pollrfd.fd = 0;
	pollrfd.events = POLLIN;
	pollrfd.revents = 0;

	ret = poll (&pollrfd, pollnfds, time);

	if (pollrfd.revents & POLLIN)
		FD_SET (0,rfd);
	return (ret);
}
#endif /*SYSV*/

/****************************************************************************\
 FUNCTION: safeopenscores.

\****************************************************************************/

int safeopenscores ()

{
    int fd;

    switch ( fd = openscore ( SCOREFILE, TIMEOUT, OPENSCORE ) ) {
    case -3:
	perror ( SCOREFILE );
	exit ( 1 );
	break;
    case -2:
	printf ( "Creating score file: %s\n", SCOREFILE );
	switch ( newscore ( SCOREFILE, TIMEOUT ) ) {
	case -2:
	    printf ( "Score file doesnt really exist?\n" );
	    perror ( SCOREFILE );
	    exit ( 2 );
	    break;
	case -1:
	    printf ( "Score file does exist but is locked!\n" );
	    perror ( SCOREFILE );
	    exit ( 3 );
	    break;
	}
	exit ( 0 );
	break;
    case -1:
	printf ( "Score file does exist but is locked!\n" );
	perror ( SCOREFILE );
	exit ( 4 );
	break;
    }
    return ( fd );
}

/****************************************************************************\
 FUNCTION: printscores.

\****************************************************************************/

void printscores ( option )

char *option;

{
    int num, dummy, fd, i;

    num = width ( option+2, DEFSCORES, & dummy );
    if ( num < 1 || num > STSIZE ) {
	printf ( "Phlegm off - thats a silly number!\n" );
	exit ( 1 );
    }
    fd = safeopenscores ();
    (void) printf ( ".-----.---------------------------.-----------.\n");
    (void) printf ( "| No. | Name                      | Score     |\n");
    (void) printf ( ">-----+---------------------------+-----------<\n");
    for ( i = 1; i <= num; i++ )
	(void) printf ( "| %3d %s\n",
		       i,
		       getscore ( fd, i, "| %-25n | %9s |" ) );

    (void) printf ( "`-----^---------------------------^-----------'\n");
    (void) closescore ( fd );
}

/****************************************************************************\
 PROCEDURE: dohiscore.

\****************************************************************************/

void dohiscore ( score )

int score;

{
    int fd, rank;
    char *name, *getenv ();

    fd = safeopenscores ();
    if (   ( name = getenv ( "MONGERNAME" ) ) == NULL
	&& ( name = getenv ( "NAME" ) ) == NULL
	&& ( name = getenv ( "LOGNAME" ) ) == NULL
	&& ( name = getenv ( "USER" ) ) == NULL
	&& ( name = getlogin () ) == NULL )
	    name = "Fish-monger-trainee";

    switch ( ( rank = addscore ( fd, name, score ) ) ) {
    case -1:
	perror ( "Cheese up in score file! ");
	exit ( 1 );
	break;
    case 0:
	break;
    default:
	(void) move ( 10, 22 );
	(void) printw ( "You rank #%d in the hiscore table!\n", rank );
	(void) refresh ();
	(void) usleep ( 500000 );
	break;
    }
    (void) closescore ( fd );
}

/****************************************************************************\
 FUNCTION: pollkeyboard.
 Finds out whether there are any characters to be read in from stdin.
 It returns the following values:-
       0  : No char to be read in the past (sec+usec/10^6).
       1  : There are chars to be read.
      -1  : Phlegm up!

\****************************************************************************/

int pollkeyboard ( sec, usec )

int sec, usec;

{

    fd_set rd, wr, xp;
    struct timeval timeout;

    timeout.tv_sec = sec;
    timeout.tv_usec = usec;

    FD_ZERO ( &rd );
    FD_ZERO ( &wr );
    FD_ZERO ( &xp );

    FD_SET ( 0, &rd );

    switch ( select ( 32, &rd, &wr, &xp, &timeout ) ) {
    case -1:
	return ( -1 );
    case 0:
	return ( 0 );
    default:
	if ( FD_ISSET ( 0, &rd ) )
	    return ( 1 );
	else
	    return ( 0 );
    }
}

/***************************************************************************\
|*									   *|
|*  The function flush_keybord() checks to see whether there are any	   *|
|*  keystrokes waiting to be read, and if so reads them and throws	   *|
|*  them away.	Surprising how complex this needs to be, but the only	   *|
|*  alternative I can think of is to ioctl() stdin to be non-blocking,	   *|
|*  read() until it returns 0, then ioctl() it back, which is scarcely	   *|
|*  an improvement. Thanx to Unkle Chicken Mirk    			   *|
|*									   *|
\***************************************************************************/

void flush_keyboard ()
{
  (void) ioctl ( fileno(stdin), TCFLSH, 0 );	
	
#if 0
  fd_set fds;			/* Will contain only stdin */
  struct timeval timeout;	/* Will contain a zero timeout (poll) */
  static char buffer[NUM_COLUMNS];	/* Will be used as the buffer to read(2) */
  int status;			/* Contains select(2)'s return value */

  while (1) {
    FD_ZERO (&fds);
    FD_SET (0, &fds);
    timeout.tv_sec = 0L;
    timeout.tv_usec = 0L;
    if (((status = select (1, &fds, (fd_set*) NULL, (fd_set*) NULL,
			   &timeout)) == -1) && (errno != EINTR))
      exit ( 1 );
    if ((status == 0) || (status == -1))
      break;			/* Can only be -1 if errno == EINTR */
    (void) read (0, buffer, NUM_COLUMNS);
  }
#endif
}

/****************************************************************************\
 FUNCTION: checkoption.
 Receives a string, and checks to see if string has been passed as an argument
 to the program.
 Returns position if found, 0 if not.

\****************************************************************************/

int checkoption( numoptions, option, checkoption )

    char *checkoption, *option[];
    int numoptions;

{
    int i;

    for ( i=1 ; i<numoptions ; i++ )
	if ( ! strncmp ( option[i] , checkoption, strlen ( checkoption ) ) )
	    return ( i );
    return ( 0 );
}

/****************************************************************************\
 FUNCTION: check_load_average

\****************************************************************************/
 
int check_load_average()

{
    int load_average_ok = 0;

    load_average_ok = system ( "/usr/local/gamesok -s" );

    return( load_average_ok );
}

/****************************************************************************\
 FUNCTION: rnd.  Returns a random integer.

\****************************************************************************/

int rnd( range )

int range;

{
    return ( random() % range );
}

/****************************************************************************\
  FUNCTION: safe_position

\****************************************************************************/

int safe_position()

{
    int t, count = 0;
    int ch;

    for ( t=STARTING_X ; t<NUM_COLUMNS-5 ; t++ ) {
	(void) move( ship_y, t );
	if ( ISBACK (inch ()) )
		count++;
	else
		count = 0;
	if ( count == 10 ) break;
    }
    return( t );
	
}


/****************************************************************************\
  FUNCTION: active_death

\****************************************************************************/

void active_death( index )

    int index;
{
    if ( on_screen( active[index].x_pos ) == ON ) {
	(void) move( active[index].y_pos, active[index].x_pos );
	(void) addstr ( blank_things[active[index].type] );
    }

    active[index].type = UNUSED;
    active[index].x_pos = ELSEWHERE;
    
    if ( index == array_top-1 ) 
	while ( active[array_top-1].type == UNUSED )
	    array_top--;
}

/****************************************************************************\
  FUNCTION: active_birth

\****************************************************************************/

void active_birth( term, new_type, new_x, new_y, new_info )

    int term;
    int new_type;
    int new_x;
    int new_y;
    int new_info;
{
    int new_index = 0;

    if ( term == LONG_TERM ) {
	while ( active[new_index].type != UNUSED )
	    new_index++;
	if ( new_index == array_top )
	    array_top++;
    } else {
	new_index = array_top;
	array_top++;
    }

    active[new_index].type  = new_type;
    active[new_index].x_pos = new_x;
    active[new_index].y_pos = new_y;
    active[new_index].info  = new_info;

}

/****************************************************************************\
  FUNCTION: small_explosion

\****************************************************************************/

void small_explosion( x, y )

    int x, y;
{
    int t;
    
    for ( t=0 ; t<DEBRIS_BITS/2 ; t++ )
	active_birth( SHORT_TERM, DEBRIS1, x, y, SMALL_EXPLOSION_SIZE );
    
    for ( t=0 ; t<DEBRIS_BITS/2 ; t++ )
	active_birth( SHORT_TERM, DEBRIS2, x, y, SMALL_EXPLOSION_SIZE );
    
}


/****************************************************************************\
  FUNCTION: clear_radar

\****************************************************************************/

void clear_radar( index )

    int index;

{
    int u, v;

    if ( active[index].type != LANDER ) return;

    v = 1+(active[index].y_pos-5)/6;
    u = 38+(active[index].x_pos)/80;
    if ( u>26 && u<49 ) {
	(void) move( v, u );
	(void) addch( ' ' );
    }
}

/****************************************************************************\
  FUNCTION: put_humanoids.

\****************************************************************************/

void put_humanoids()

{
    int t;

    for ( t=0 ; t<NUM_HUMANOIDS ; t++ ) {
	(void) move( 1, NUM_COLUMNS-t-11 );
	(void) addch( 'P' );
    }
    for ( t=0 ; t<NUM_HUMANOIDS-num_humanoids ; t++ ) {
	(void) move( 1, NUM_COLUMNS-t-11 );
	(void) addch( ' ' );
    }
    (void) refresh();
}

/****************************************************************************\
  FUNCTION: pod_burst

\****************************************************************************/

void pod_burst( index )

int index;

{
    int t;

    for ( t=0 ; t<NUM_SWARMERS ; t++ ) {
	active_birth( LONG_TERM, 
		     SWARMER, 
		     active[index].x_pos,
		     active[index].y_pos,
		     0 );
    }	
}

/****************************************************************************\
  FUNCTION: put_smart_bombs.

\****************************************************************************/

void put_smart_bombs()

{
    int t;

    for ( t=0 ; t<NUM_SMART_BOMBS ; t++ ) {
	(void) move( 3, NUM_COLUMNS-3*t-10 );
	(void) addstr( ">O" );
    }
    for ( t=0 ; t<NUM_SMART_BOMBS-smart_bombs ; t++ ) {
	(void) move( 3, NUM_COLUMNS-3*t-10 );
	(void) addstr( "  " );
    }
}

/****************************************************************************\
  FUNCTION: put_score.

\****************************************************************************/

void put_score()

{
    (void) move( 1, 9 );
    (void) printw( "%6i", score );

    if ( score > next_smart_bomb ) {
	if ( smart_bombs < 3 ) smart_bombs++;
	put_smart_bombs();
	next_smart_bomb += NEW_SMART_BOMB;
    }
	
}

/****************************************************************************\
  FUNCTION: laserhit

\****************************************************************************/

void laserhit( where_x, where_y )

    int where_x, where_y;

{
    int t;

    for ( t=0 ; t<array_top ; t++ ) 
	if ( active[t].type != UNUSED &&
	    active[t].y_pos == where_y &&
	    abs(active[t].x_pos-where_x) < 3 ) {
	    clear_radar( t );
	    switch ( active[t].type ) {
	    case LANDER:
		small_explosion( active[t].x_pos, active[t].y_pos );
		active_death( t );
		score += 50;
		break;
	    case MUTANT:
		small_explosion( active[t].x_pos, active[t].y_pos );
		active_death( t );
		score += 300;
		break;
	    case SWARMER:
		small_explosion( active[t].x_pos, active[t].y_pos );
		active_death( t );
		score += 25;
		break;
	    case BOMBER:
		small_explosion( active[t].x_pos, active[t].y_pos );
		active_death( t );
		score += 100;
		break;
	    case BAITER:
		small_explosion( active[t].x_pos, active[t].y_pos );
		active_death( t );
		score += 400;
		break;
	    case POD:
		pod_burst( t );
		active_death( t );
		score += 500;
		break;
	    case HUMANOID:
		small_explosion( active[t].x_pos, active[t].y_pos );
		if ( active[t].info != HUMANOID_FREE ) 
		    active[active[t].info].info = LANDER_IDLE;
		active_death( t );
		num_humanoids--;
		put_humanoids();
		if ( num_humanoids == 0 ) {
		    hit += 1;
		    lives = 1;
		}
		score -= 1000;
		break;
	    case MISSILE:
		active_death( t );
		score += 5;
		break;
	    case BOMB:
	    default:
		break;
	    }
	    put_score();
	    break;
	}
    
}

/****************************************************************************\
  FUNCTION: clean_up

\****************************************************************************/

void clean_up()

{
    int t;
    
    for ( t=0 ; t<array_top ; t++ ) {
	switch ( active[t].type ) {
	case LANDER:
	case MUTANT:
	case BAITER:
	case POD:
	case SWARMER:
	case HUMANOID:
	    break;
	default:
	    active_death( t );
	    break;
	}
    }
}

/****************************************************************************\
 PROCEDURE: ship_die. Blow the ship all over the shop.

\****************************************************************************/

void ship_die ()

{
    int count = 0, old_ship_x, t, tme = 0;
    float bx[16], by[16], vx[16], vy[16];
    char bits[17];
    int flag[16];

    laserhit( ship_x, ship_y );
    old_ship_x = safe_position();
    clean_up();

    (void) move ( ship_y, ship_x );
    (void) addstr ( "    " );
    (void) strcpy ( bits , ">C=-++--++--''``" );
    for ( t=0 ; t<16 ; t++ ) {
	flag[t] = 0;
	bx[t] = ship_x;
 	by[t] = ship_y;
 	vx[t] = ( (float) (rnd(10000))/3700.0 + 0.06) ;
 	if ( rnd(2) == 0 ) vx[t] = -vx[t];
 	vy[t] = ( (float) (rnd(10000))/10005.0 + 0.02);
 	if ( rnd(2) == 0 ) vy[t] = -vy[t];
    }
    while ( count < 16 && tme < 50 ) {
	tme++;
	for ( t=0 ; t<16 ; t++ ) {
	    (void) move ( (int) (by[t]), (int) (bx[t]) );
	    (void) addch ( ' ' );
	    if ( flag[t] == 0 ) {

 		bx[t] = bx[t] + vx[t];
 		if ( ( bx[t] < 1 ) || ( bx[t] > NUM_COLUMNS-1 ) ) {
		    flag[t] = 1;
		    count++;
		}
		by[t] = by[t] + vy[t];
		vy[t] += 0.04;
		if ( ( by[t] < 3 ) || ( by[t] > BOTTOM-1 ) ) {
		    flag[t] = 1;
		    count++;
		}
		(void) move ( (int) (by[t]), (int) (bx[t]) );
		(void) addch ( bits[t] );
	    }
	}
	(void) refresh();
	(void) usleep ( EXPLOSION_DELAY );
    }
    lives--;
    ship_x = old_ship_x;

}

/****************************************************************************\
 PROCEDURE: fire_laser.
 Draws the hoopy laser beam and updates the screen.

\****************************************************************************/

void fire_laser ()

{
    int t, x1, x2;
    char check;

    /* put beam on */

    if ( facing == RIGHT ) {	
	x1 = ship_x+4;
	x2 = NUM_COLUMNS-1;
	for ( t=x1+1 ; t!=x2 ; t++ ) {
	    (void) move ( ship_y, t );
	    check = inch();
	    if ( ! ISBACK (check) ) {
		laserhit ( t, ship_y );       
		break;
	    }
	    (void) addch ( LASER_BEAM_CHAR );
	}
    } else {
	x1 = 0;
	x2 = ship_x-1;
	for ( t=x2-1 ; t!=x1 ; t-- ) {
	    (void) move( ship_y, t );
	    check = inch();
	    if ( ! ISBACK (check) ) {
		laserhit ( t, ship_y );    
		break;
	    }
	    (void) addch ( LASER_BEAM_CHAR );
	}
    }

    (void) refresh();
    (void) usleep ( SMALL_DELAY );

    /* take beam off */

    for ( t=x1+1 ; t!=x2 ; t++ ) {
	(void) move ( ship_y, t );
	(void) addch ( ' ' );
    }
    (void) refresh();
}

/****************************************************************************\
  FUNCTION: turn_messages.

\****************************************************************************/

void turn_messages( mode )

int mode;

{
    if ( mode == OFF )
	(void) system ( "/bin/mesg n" );
    else
	(void) system ( "/bin/mesg y" );
}

/****************************************************************************\
  FUNCTION: remove_cursor.

\****************************************************************************/

void remove_cursor()

{
    char *termtype, *getenv();

    termtype = getenv ( "TERM" );

    if ( strncmp ( termtype, "tvi", 3 ) == 0 ) {
	(void) addstr ( ".0 " ); 
	(void) refresh();
    }
    
}

/****************************************************************************\
  FUNCTION: replace_cursor.

\****************************************************************************/

void replace_cursor()

{
    char *termtype, *getenv();

    termtype = getenv ( "TERM" );

    if ( strncmp ( termtype, "tvi", 3 ) == 0 )
	(void) addstr ( ".1 " ); 
    
    (void) printf( "\n\nGoodbye.\n\n" );
    (void) refresh();
}

/****************************************************************************\
  FUNCTION: define_keys.

\****************************************************************************/

void define_keys() 

{
    extern char *getenv ();	/* Declaration to keep lint happy. */
    char *cp;			/* Temporary pointer for getenv() */
    
    if ((cp = getenv ("MONGERKEYS")) != NULL) {
	if (*cp != '\0')
	    _up_key = *(cp++);
	if (*cp != '\0')
	    _down_key = *(cp++);
	if (*cp != '\0')
	    _right_key = *(cp++);
	if (*cp != '\0')
	    _left_key = *(cp++);
	if (*cp != '\0')
	    _fire_key = *(cp++);
	if (*cp != '\0')
	    _smart_bomb_key = *(cp++);
	if (*cp != '\0')
	    _hyperspace_key = *(cp++);
    }
}

/****************************************************************************\
  FUNCTION: put_lives.

\****************************************************************************/

void put_lives()

{
    int t;

    for ( t=0 ; t<NUM_LIVES ; t++ ) {
	(void) move( 3, 5*t+2 );
	(void) addstr( ">8=-" );
    }
    for ( t=0 ; t<NUM_LIVES-lives+1 ; t++ ) {
	(void) move( 3, 5*t+2 );
	(void) addstr( "    " );
    }
}

/****************************************************************************\
  FUNCTION: on_screen
  Used by the ALIENS to check if they should be displayed or not.

\****************************************************************************/

int on_screen( x_coord )

    int x_coord;
{
    if ( x_coord < 0 || x_coord > NUM_COLUMNS-3 ) return( OFF );
    return( ON );
}

/****************************************************************************\
  FUNCTION: in_space
  Used by bullets etc. to see if they have gone out of range

\****************************************************************************/

int in_space( y_coord, x_coord )

    int y_coord, x_coord;

{
    if ( y_coord<5 || y_coord>BOTTOM || x_coord<0 || x_coord>NUM_COLUMNS )
	return( 0 );
    else
	return( 1 );

}

/****************************************************************************\
  FUNCTION: put_radar()

\****************************************************************************/

void put_radar()

{
    int t, u, v;

    (void) move ( 1, 27 );
    (void) addstr( "                      " );
    (void) move ( 2, 27 );
    (void) addstr( "           _          " );
    (void) move ( 3, 27 );
    (void) addstr( "                      " );

    for ( t=0 ; t<array_top ; t++ ) {
	if ( active[t].type == LANDER ) {
	    v = 1+(active[t].y_pos-5)/6;
	    u = 38+(active[t].x_pos)/80;
	    if ( u>26 && u<49 ) {
		(void) move( v, u );
		if ( active[t].info == LANDER_IDLE ) 
		    (void) addch( '.' );
		else 
		    (void) addch( ',' );
	    }
	}
    }
}

/****************************************************************************\
  FUNCTION: draw_screen_detail.

\****************************************************************************/

void draw_screen_detail()

{
    (void) clear();

    (void) move( 0, 0 );
    (void) addstr( "                           .----------v----------." );
    (void) move( 1, 0 );
    (void) addstr( "                          /                       \\" );
    (void) move( 2, 0 );
    (void) addstr( "                        -<            _            >-" );
    (void) move( 3, 0 );
    (void) addstr( "                          \\                       /" );
    (void) move( 4, 0 );
    (void) addstr( "...........................`----------^----------'............................." );

    put_lives();
    put_smart_bombs();
    put_humanoids();
    put_score();
    put_radar();

    (void) refresh();
}

/****************************************************************************\
  FUNCTION: set_conditions.

\****************************************************************************/

void set_conditions()

{
    turn_messages( OFF );
    (void) initscr();
    (void) cbreak();
    (void) noecho();
    remove_cursor();
    (void) srandom ( time(&big) );
}

/****************************************************************************\
  FUNCTION: set_variables.

\****************************************************************************/

void set_variables()

{
    int t;

    lives = NUM_LIVES;
    smart_bombs = NUM_SMART_BOMBS;

    for ( t=0 ; t<ACTIVE_ARRAY_SIZE ; t++ )
	active[t].type = UNUSED;

    array_top = 0;

    lander_decide_steal = LANDER_DECIDE_STEAL;
    num_humanoids       = NUM_HUMANOIDS;
    new_alien_rate      = NEW_ALIEN_RATE;
    next_smart_bomb     = NEW_SMART_BOMB;
    num_landers         = NUM_LANDERS;
}

/****************************************************************************\
  FUNCTION: remove_ship.

\****************************************************************************/

void remove_ship()

{
    (void) move( ship_y, ship_x );
    (void) addstr( "    " );
}

/****************************************************************************\
  FUNCTION: place_ship.
  Returns 1 if ship has hit something.

\****************************************************************************/

int place_ship()

{

    int t;
    char check;

    for ( t=0 ; t<4 ; t++ ) {
	(void) move ( ship_y, ship_x+t );
 	check = inch();
 	if ( ( check != ship[facing][t] ) && !(ISBACK (check)) )
		return 1;
	(void) addch( ship[facing][t] );
    }
    (void) refresh();
    return ( 0 );
}

/****************************************************************************\
  FUNCTION: reverse_ship

\****************************************************************************/

void reverse_ship()

{
    facing = 1-facing;
    if ( home_x == LEFT_X_POSITION )
	home_x = RIGHT_X_POSITION;
    else
	home_x = LEFT_X_POSITION;
}

/****************************************************************************\
  FUNCTION: thrust_ship_left

\****************************************************************************/

void thrust_ship_left()

{
    if ( thrust < MAX_THRUST ) 
	thrust++;
    if ( thrust > 0 && facing == RIGHT ) reverse_ship();
}

/****************************************************************************\
  FUNCTION: thrust_ship_right

\****************************************************************************/

void thrust_ship_right()

{
    if ( thrust > -MAX_THRUST ) 
	thrust--;
    if ( thrust < 0 && facing == LEFT ) reverse_ship();
}

/****************************************************************************\
  FUNCTION: move_ship_up

\****************************************************************************/

void move_ship_up()

{
    if ( ship_y < TOP ) return;
    remove_ship();
    ship_y--;
    hit += place_ship();
}

/****************************************************************************\
  FUNCTION: move_ship_down.

\****************************************************************************/

void move_ship_down()

{
    if ( ship_y > BOTTOM-1 ) return;
    remove_ship();
    ship_y++;
    hit += place_ship();

}

/****************************************************************************\
  FUNCTION: dematerialize_ship

\****************************************************************************/

void dematerialize_ship()

{
    int t, zx, zy, vx, vy;

    for ( t=0 ; t<NUM_ZOOMS; t++ ) {
	zx = ship_x+rnd(4);
	zy = ship_y;
	vx = rnd(4)+1;
	if ( rnd(2) == 0 ) vx =-vx;
	vy = rnd(3)+1;
	if ( rnd(2) == 0 ) vy =-vy;

	while ( zx>4 && zx<NUM_COLUMNS-4 && zy>3 && zy<BOTTOM ) {
	    (void) move ( zy, zx );
	    (void) addch ( '.' );
	    (void) refresh();
	    zx += vx;
	    zy += vy;
	}
    }
}

/****************************************************************************\
  FUNCTION: materialize_ship

\****************************************************************************/

void materialize_ship()

{
    int t, count = 0;

    static int _bit_x[8] = { 0, 0, NUM_COLUMNS/3-1, NUM_COLUMNS/3, 
	(NUM_COLUMNS*2/3), (NUM_COLUMNS*2/3)-1, NUM_COLUMNS-1, NUM_COLUMNS-2 };
    static int _bit_y[8] = { 6, BOTTOM-1, 6, BOTTOM-1, 
			    6, BOTTOM-1, 6, BOTTOM-1 };

    int bit_x[8], bit_y[8];

    for ( t=0 ; t<8 ; t++ ) {
	bit_x[t] = _bit_x[t];
	bit_y[t] = _bit_y[t];
    }    

    while ( count != 16 ) {
	count = 0;
	for ( t=0 ; t<8 ; t++ ) {
	    (void) move( bit_y[t], bit_x[t] );
	    (void) addch( ' ' );
	    if ( bit_x[t] > ship_x+(t/2) ) bit_x[t]--;
	    else if ( bit_x[t] < ship_x+(t/2) ) bit_x[t]++;
	    else count++;
	    if ( bit_y[t] > ship_y && rnd(6) == 0  ) bit_y[t]--;
	    else if ( bit_y[t] < ship_y && rnd(6) == 0  ) bit_y[t]++;
	    else if ( bit_y[t] == ship_y ) count++;
	    (void) move( bit_y[t], bit_x[t] );
	    (void) addch( pieces[facing][t] );
	}
	(void) refresh();
	(void) usleep( MATERIALIZE_DELAY );
    }
}

/****************************************************************************\
  FUNCTION: hyperspace

\****************************************************************************/

void hyperspace()
    
{
    int t, shift;
    
    dematerialize_ship();
    
    shift = rnd(SPACE_RANGE)*(2*facing-1);
    for ( t=0 ; t<array_top ; t++ ) {
	if ( active[t].type != UNUSED ) 
	    active[t].x_pos += shift;
    }

    draw_screen_detail();
    materialize_ship();

}

/****************************************************************************\
  FUNCTION: start_aliens

\****************************************************************************/

void start_aliens()

{
    int i;

    for ( i=0 ; i<NUM_HUMANOIDS ; i++ ) {
	active_birth( LONG_TERM,
		     HUMANOID,
		     rnd(2*SPACE_RANGE)-SPACE_RANGE, 
		     BOTTOM,
		     HUMANOID_FREE );
    }

    for ( i=0 ; i<NUM_BOMBERS ; i++ ) {
	active_birth( LONG_TERM,
		     BOMBER,
		     rnd(2*SPACE_RANGE)-SPACE_RANGE, 
		     rnd(BOTTOM-10)+7,
		     1 );
    }

    for ( i=0 ; i<NUM_PODS ; i++ ) {
	active_birth( LONG_TERM,
		     POD,
		     rnd(2*SPACE_RANGE)-SPACE_RANGE, 
		     rnd(BOTTOM-10)+7,
		     0 );
    }

    for ( i=0 ; i<num_landers ; i++ ) {
	active_birth( LONG_TERM,
		     LANDER,
		     rnd(2*SPACE_RANGE)-SPACE_RANGE, 
		     rnd(BOTTOM-10)+7,
		     LANDER_IDLE );
    }
    
}

/****************************************************************************\
  FUNCTION: beep

\****************************************************************************/
#ifndef SYSV
void beep()

{
    (void) move( 0, 0 );
    (void) addstr( BEEP_CHARS );
    (void) refresh();
    
}
#endif

/****************************************************************************\
  FUNCTION: start_life

\****************************************************************************/

void start_life()

{
    draw_screen_detail();

    ship_x = STARTING_X;   
    ship_y = STARTING_Y;
    home_x = STARTING_X;
    facing = RIGHT;
    thrust = -1;
    hit = 0;

    materialize_ship();
}

void move_swarmer( index )

int index;

{
    active[index].x_pos += rnd(3)-1;

    if ( rnd(2) == 0 && active[index].y_pos > TOP ) 
	active[index].y_pos--;
    else if ( active[index].y_pos < BOTTOM-1 ) 
	active[index].y_pos++;
}

void move_baiter( index )

int index;

{
    active[index].x_pos += active[index].info;
    if ( rnd(BAITER_TURNAROUND) == 0 )
	active[index].info = -active[index].info;
    
    if ( rnd(2) == 0 && active[index].y_pos > TOP ) 
	active[index].y_pos--;
    else if ( active[index].y_pos < BOTTOM-1 ) 
	active[index].y_pos++;
}

void move_mutant( index )

int index;

{
    if ( active[index].x_pos > ship_x )
	active[index].x_pos--;
    else
	active[index].x_pos++;

    if ( active[index].y_pos > ship_y )
	active[index].y_pos--;
    else
	active[index].y_pos++;
    
    active[index].x_pos += rnd(3)-1;
    
}

void move_lander( index )

int index;

{
    int u, v;
    char radar_char = '.';

    if ( (on_screen( active[index].x_pos ) == ON ) &&
	( rnd(LANDER_FIRE_RATE) == 0 ) ) {
	    if ( ship_x > active[index].x_pos )
		active_birth( SHORT_TERM,
			     MISSILE,
			     active[index].x_pos,
			     active[index].y_pos,
			     2 );
	    else
		active_birth( SHORT_TERM,
			     MISSILE,
			     active[index].x_pos,
			     active[index].y_pos,
			     -2 );
	}
    
    switch ( active[index].info ) {
    case LANDER_IDLE:
	if ( rnd(LANDER_MATCH_HEIGHT) == 0 ) {
	    if ( active[index].y_pos > ship_y )
		active[index].y_pos--;
	    else
		active[index].y_pos++;
	}
	active[index].x_pos += rnd(3)-1;
	if ( rnd(lander_decide_steal) == 0 ) {
	    /* decide to steal */
	    beep();
	    active[index].info = 0;
	    while ( active[active[index].info].type != HUMANOID &&
		    active[index].info<array_top )
		active[index].info++;
	}
	break;
    case LANDER_STEALING:
	radar_char = ';';
	if ( rnd(LANDER_VERTICAL_RATE) == 0 ) 
	    if ( (active[index].y_pos--) == TOP ) {
		active[index].type = MUTANT;
		radar_char = ' ';
	    }
	break;
    default:
	/* tracking */
	if ( active[index].x_pos > active[active[index].info].x_pos )
	    active[index].x_pos--;
	else if ( active[index].x_pos < active[active[index].info].x_pos )
	    active[index].x_pos++;
	else {
	    if ( rnd( LANDER_VERTICAL_RATE ) == 0 )
		active[index].y_pos++;
	    
	    if ( BOTTOM == active[index].y_pos+1 ) {
		/* pick up humanoid */
		active[active[index].info].info = index;
		active[index].info = LANDER_STEALING;
	    }
	}
	radar_char = ',';
	break;
    }

    v = 1+(active[index].y_pos-5)/6;
    u = 38+(active[index].x_pos)/80;
    if ( u>26 && u<49 ) {
	(void) move( v, u );
	(void) addch( radar_char );
    }
}

void move_debris( index )

int index;

{
    active[index].y_pos += rnd(3)-1;
    active[index].x_pos += rnd(7)-3;
    if ( (active[index].info--) == 0 )
	active_death( index );
}

void move_humanoid( index )

int index;

{
    if ( active[index].info == HUMANOID_FREE ) {
	if ( active[index].y_pos == BOTTOM ) return;
	active[index].y_pos++;
	score += 75;
	put_score();
	return;
    }

    switch ( active[active[index].info].type ) {
    case LANDER:
	active[index].x_pos = active[active[index].info].x_pos+2;
	active[index].y_pos = active[active[index].info].y_pos+1;	
	break;
    case MUTANT:
	active_death( index );
	num_humanoids--;
	put_humanoids();
	if ( num_humanoids == 0 ) {
	    hit += 1;
	    lives = 1;
	}
	break;
    default:
	active[index].info = HUMANOID_FREE;
	break;
    }
}

void move_missile( index )

int index;

{
    active[index].x_pos += active[index].info;
    
    if ( in_space( active[index].y_pos, active[index].x_pos ) == 0 )
	    active_death( index );
}

void move_bomb( index )

int index;

{
    if ( in_space( active[index].y_pos, active[index].x_pos ) == 0 )
	active_death( index );
}

void move_bomber( index )

int index;

{
    if ( rnd(BOMBER_FOLLOW_DELAY) == 0 ) {
	if ( active[index].x_pos < 40 )
	    active[index].x_pos++;
	else 
	    active[index].x_pos--;
    } else 
	active[index].x_pos += rnd(3)-1;

    if ( rnd(BOMBER_DELAY) == 0 ) {
	active[index].y_pos += active[index].info;
	if ( active[index].y_pos == BOTTOM-1 || active[index].y_pos == 5 )
	    active[index].info = -active[index].info;
    }
    if ( on_screen( active[index].x_pos ) && rnd(BOMB_FREQUENCY) == 0 ) {
	active_birth( SHORT_TERM,
		     BOMB,
		     active[index].x_pos,
		     active[index].y_pos,
		     0 );
    }
}

#ifdef DO_GROUND
void init_ground ()
{
	int g;
	int h;
	int s;
	int t;

#define slope(h,t)	((h) == (t) ? 0 : ((h) < (t) ? 1 : -1))

	h = 0;
	t = rnd(MAX_GROUND);
	s = slope (h, t);

	for (g = 0; g < NUM_GROUND; g++) {
		switch (s) {
		case -1:
			ground_height[g] = --h;
			ground_slope[g] = -1;
			break;

		case 0:
			ground_height[g] = h;
			ground_slope[g] = 0;
			break;

		case 1:
			ground_height[g] = h++;
			ground_slope[g] = 1;
			break;
		}

		if (h == t) {
			t = rnd(MAX_GROUND);
			s = slope (h, t);
		}
	}
}

void do_ground (int pos, char *draw)
{
	static int g = 0;
	int ng;
	int i;
	int d;

	g -= pos;
	if (g < 0)
		g += (NUM_GROUND * GROUND_DIV);
	g %= (NUM_GROUND * GROUND_DIV);

	ng = g / GROUND_DIV;

	for (i = 0; i < NUM_COLUMNS; i++) {
		move (BOTTOM - 1 - ground_height[ng], i);
		addch (draw[ground_slope[ng]+1]);

		ng++;
		ng %= NUM_GROUND;
	}
}
#endif /* DO_GROUND */

/****************************************************************************\
  FUNCTION: move_things.

\****************************************************************************/

void move_things()

{
    int slide_window = 0;
    int t;

#ifdef DO_GROUND
    if (ground)
	    do_ground (0, "   ");
#endif

    if ( ship_x != home_x ) {
	if ( ship_x > home_x )
	    slide_window = -1+(home_x-ship_x)/WINDOW_SLIDE_FACTOR;
	else
	    slide_window =  1+(home_x-ship_x)/WINDOW_SLIDE_FACTOR;

	remove_ship();
	ship_x += slide_window;
    }

#ifdef DO_GROUND
    if (ground)
	    do_ground (thrust + slide_window, "\\_/");
#endif

    for ( t=0 ; t<array_top ; t++ ) {
	if ( active[t].type != UNUSED ) {

	    if ( on_screen( active[t].x_pos ) == ON ) {
		(void) move( active[t].y_pos, active[t].x_pos );
		(void) addstr( blank_things[active[t].type] );
	    }
	    clear_radar( t );

	    active[t].x_pos += slide_window + thrust;

	    switch( active[t].type ) {
	    case LANDER:
		move_lander( t );
		break;
	    case MUTANT:
		move_mutant( t );
		break;
	    case BAITER:
		move_baiter( t );
		break;
	    case BOMBER:
		move_bomber( t );
		break;
	    case SWARMER:
		move_swarmer( t );
		break;
	    case HUMANOID:
		move_humanoid( t );
		break;
	    case BOMB:
		move_bomb( t );
		break;
	    case MISSILE:
		move_missile( t );
		break;
	    case DEBRIS1:
	    case DEBRIS2:
		move_debris( t );
		break;
	    case POD:
	    default:
		break;
	    }
	    if ( on_screen( active[t].x_pos ) == ON ) {
		(void) move( active[t].y_pos, active[t].x_pos );
		(void) addstr( things[active[t].type] );
	    }
	}
    }
    
    hit += place_ship();
}

/****************************************************************************\
  FUNCTION: introduction.

\****************************************************************************/

void introduction()

{

    int t;
    static char screen[20][80] = { "    _____ ____ ____	       ______  _____  ___     _____  ____ _____   ",
				       "   /       /  /      /    /    \\  \\  \\ \\    \\ \\  \\  \\ \\      \\    \\    \\  ",
				       "  /___    /  /____  /____/      \\  \\  \\ \\    \\ \\  \\  \\ \\  ___ \\___ \\____\\ ",
				       " /       /       / /    /        \\  \\  \\ \\    \\ \\  \\  \\ \\    \\ \\    \\  \\_ ",
				       "/     __/__ ____/ /    /          \\     \\ \\____\\ \\  \\__\\ \\____\\ \\____\\   \\",
				       " ",
				       "	      >o<              	}X{                  *   ",
				       "       	     Lander            Mutant             Swarmer ",
				       " ",
				       "        P       	     		                      x ",
				       "    Humanoid                    >8=-                         Bomb  ",
				       "	                                               	  ",
				       "       	       %       	       	-=-    	       	     ()	  ",
				       "	     Bomber            Baiter               Pod   ",
				       "						       ",
				       "_____ ____ ____	          	   ______  _____  ___     _____  ____ _____",
				       "\\   	\\  \\   	 \\    \\   	  /  /  / /    / /  /  / /      /    /    /",
				       " \\___    \\  \\____ \\____\\         /  /  / /    / /  /  / / ___  /___ /____/ ",
				       "  \\ 	  \\      \\ \\	\\       /     / /    / /  /  / /    / /    /  /   ",
				       "   \\     __\\__ ___\\ \\    \\     /     / /____/ /  /__/ /____/ /___ /   / " };
    


    (void) clear();
    
    for ( t=0 ; t<10 ; t++ ) {
	(void) move ( t, 0 );
	(void) addstr( screen[t] );
	(void) move ( 19-t, 0 );
	(void) addstr( screen[19-t] );
	(void) refresh();
	(void) usleep( 150000 );
    }
    (void) sleep( 3 );
    for ( t=0 ; t<10 ; t++ ) {
	(void) move ( t, 0 );
	(void) addstr( "                                                                              " );
	(void) move ( 19-t, 0 );
	(void) addstr( "                                                                              " );
	(void) refresh();
	(void) usleep( 150000 );
    }    
}

/****************************************************************************\
  FUNCTION: energize_smart_bomb

\****************************************************************************/

void energize_smart_bomb()

{
    int t;
    
    if ( smart_bombs == 0 ) return;
    smart_bombs--;
    put_smart_bombs();
    
    for ( t=0 ; t<array_top ; t++ ) {
	switch( active[t].type ) {
	case LANDER:
	case MUTANT:
	case SWARMER:
	case BAITER:
	case BOMBER:
	case POD:
	    if ( on_screen( active[t].x_pos ) == ON )
		laserhit( active[t].x_pos, active[t].y_pos );
	    break;
	default:
	    break;
	}
    }
}

/****************************************************************************\
  FUNCTION: new_aliens

\****************************************************************************/

void new_aliens()

{
    int choice, what, infr, t;
    
    if ( array_top > NO_MORE_ACTIVITY || rnd(new_alien_rate) > 0 ) return;

    if ( array_top < NEW_WAVE ) { 
	score += num_humanoids*100;
	for ( t=0 ; t<array_top ; t++ ) 
	    active_death( t );
	
	num_landers += 5;
	new_alien_rate -= NEW_WAVE_ALIEN_RATE_CHANGE;
	if ( new_alien_rate < 0 ) new_alien_rate = 0;
	lander_decide_steal -= NEW_WAVE_STEAL_RATE_CHANGE;
	if ( lander_decide_steal < 0 ) lander_decide_steal = 0;
	start_aliens();
    } else {
	choice = rnd(100);
	
	if ( choice< POD_CHANCE ) what = POD;
	else if ( choice< BOMBER_CHANCE ) {
	    what = BOMBER;
	    infr = 1;
	} else if ( choice< BAITER_CHANCE ) {
	    what = BAITER;
	    infr = 2;
	} else {
	    what = LANDER;
	    infr = LANDER_IDLE;
	}
	
	active_birth( LONG_TERM,
		     what,
		     rnd(NUM_COLUMNS),
		     rnd(BOTTOM-TOP)+TOP,
		     infr );
    }
}


/****************************************************************************\
MAIN PROGRAM. 

\****************************************************************************/

main( argc, argv )

int argc; char *argv[];

{
    char key;
    int keyflag, scoreopt;

    if ( scoreopt = checkoption( argc, argv, "-s" ) ) {
	/* print the hi-score table rather than running */
	printscores ( argv[scoreopt] );
	exit ( 0 );
    }

    define_keys();
    set_variables();
    set_conditions();

    if ( checkoption( argc, argv, "-f" ) == 0 )
	introduction();

#ifdef DO_GROUND
    if ( checkoption ( argc, argv, "-g" ) )
	    ground = 0;
    else
	    init_ground();
#endif

    start_aliens();

    while ( lives != 0 ) {
	start_life();
	while ( hit == 0 ) {

	    switch ( pollkeyboard ( 0, 50000 ) ) {
	    case -1:
		(void) printw ( "Phlegm up! ERROR 1" );
		exit ( 1 );
		break;
	    case 0 :
		/* small delay, no input from keyboard */
		(void) usleep ( SMALL_DELAY );
		keyflag = 0;
		break;
	    default:
		keyflag = 1;
		if ( read ( 0, &key, 1 ) != 1 ) {
		    (void) printw ( "Phlegm up! ERROR 2" );
		    exit ( 1 );
		}
		flush_keyboard(); 
	    }
	    
	    if ( keyflag ) {
		if (key == _up_key) {
		    move_ship_up();

		} else if (key == _down_key) {
		    move_ship_down();

		} else if (key == _right_key) {
		    thrust_ship_right();

		} else if (key == _left_key) {
		    thrust_ship_left();

		} else if (key == _fire_key) {
		    fire_laser();

		} else if (key == _smart_bomb_key) {
		    energize_smart_bomb();

		} else if ( key == _hyperspace_key ) {
		    hyperspace();
	        } else if ( key == _suicide_key ) {
		    move (10, 21);
		    addstr ("Self destruct sequence initiated ...");
		    lives = 1;
		    hit = 1;
		} else
		    (void) usleep ( SMALL_DELAY );
	    }
	    
	    if (thrust) {
		    if (++thrust_timer % THRUST_DECAY == 0)
			    thrust += (thrust > 0 ? -1 : 1);
	    } else {
		    thrust_timer = 0;
	    }

	    move_things();
	    new_aliens();

	}
	ship_die();
    }
    /* Game is dead */
    
    dohiscore( score );
    flush_keyboard();
    (void) nocbreak();
    replace_cursor();
    turn_messages( ON );
    (void) endwin();
}

